#ifndef YERBACON_PARSECOMPONENTS_HPP
#define YERBACON_PARSECOMPONENTS_HPP

#include <iostream>
#include <memory>
#include <utility>
#include <vector>
#include <typeinfo>
#include <memory>
#include <optional>
#include <ranges>
using namespace std;

#include "../lex.hpp"

#define IS(X) template<typename T, typename std::enable_if<std::is_base_of<X, T>::value>::type* = nullptr>
struct ParseComponent {
    [[nodiscard]] const type_info& getId() const { return typeid(*this); }
    virtual ~ParseComponent() = default;
};

namespace StandardComponents {
    struct NamedIdentifier: public ParseComponent {
        const string name;
        explicit NamedIdentifier(string_view nameText): name(nameText) {}
    };
    struct Define: NamedIdentifier {
        const bool final;
        explicit Define(const bool& isFinal, string_view nameText): NamedIdentifier(nameText), final(isFinal) {}
        explicit Define(string_view nameText): Define(false, nameText) {}
    };
    struct Reference: NamedIdentifier {
        using NamedIdentifier::NamedIdentifier;
    };

    struct Class: NamedIdentifier {
        struct Constructor {};
        Constructor constructor;
        using NamedIdentifier::NamedIdentifier;
    };
    namespace types {
        struct String: ParseComponent {
            const string content;
            explicit String(const char* string): content(string) {}
        };
    }
}

#define IS_PARSECOMPONENT IS(ParseComponent)
class ParseTree {
protected:
    mutable vector<unique_ptr<ParseComponent>> subComponents;
    using array_type = decltype(subComponents);
    using iterator = array_type::iterator;
    using const_iterator = array_type::const_iterator;
public:
    inline iterator begin() const noexcept { return subComponents.begin(); }
    inline const_iterator cbegin() const noexcept { return subComponents.cbegin(); }
    inline iterator end() const noexcept { return subComponents.end(); }
    inline const_iterator cend() const noexcept { return subComponents.cend(); }
    IS_PARSECOMPONENT
    auto findById() const {
        return subComponents | views::filter([](unique_ptr<ParseComponent>& it) {
            return it->getId() == typeid(T);
        }) | views::transform([](unique_ptr<ParseComponent>& it) {
            return reinterpret_cast<T*>(it.get());
        });
    }
    IS(StandardComponents::NamedIdentifier)
    optional<reference_wrapper<T>> findReferenceByName(const string& name) const {
        auto identifiers = findById<T>() | views::transform([](T* it) {
            return ref(static_cast<T&>(*it));
        });
        for (const auto& identifier: identifiers) {
            if (identifier.get().getId() == typeid(T) && identifier.get().name == name) {
                return make_optional(identifier);
            }
        }
        return optional<reference_wrapper<T>>();
    };
    inline size_t getCompCount() const { return subComponents.size(); }
    IS_PARSECOMPONENT
    void add(const T& component) const {
        subComponents.emplace_back(new T(component));
    }; IS_PARSECOMPONENT void addAll(const initializer_list<T>& components) const {
        for (const auto& comp: components) add<T>(comp);
    }
    IS_PARSECOMPONENT
    inline const ParseTree& operator<<(const T& component) const { add(component); return *this; }
    ParseTree(): subComponents() {};
    IS_PARSECOMPONENT
    explicit ParseTree(const T& element): ParseTree() { add(element); }
    IS_PARSECOMPONENT
    ParseTree(const initializer_list<T>& elements): ParseTree() { addAll(elements); }
};
#undef IS_PARSECOMPONENT

#endif //YERBACON_PARSECOMPONENTS_HPP